% This script generates estimates of the mean and variance trajectories of
% ODE-based models with polynomial rates using the Extended Kalman Filter (EKF) 
% and "Naive" Monte Carlo (NMC)

% User can choose between three systems described in the Examples section
% of the paper: 
% a bifunctional bacterial two-component system;
% a monofunctional bacterial two-component system;
% a mitogen-activated protein kinase cascade.

% !!! In order to run this code, the installation of IQMtools is required.
% (https://iqmtools.intiquan.com/).

% ! The simulation part of the code makes use of the Parallel Computing Toolbox of MATLAB.

% The code was tested on MATLAB R2018b with IQM Tools V1.2.1.

clear;
clear model_MEX moments_MEX
clc;

%----------------------------------------------------------------------------------------------------------------------------------------------------------
%%% USER-DEFINED INPUTS
nModel = 3;
% nModel denotes the model to be used:
% 1 - a bifunctional bacterial two-component system
% 2 - a monofunctional bacterial two-component system
% 3 - a mitogen-activated protein kinase cascade

momentEqs = 1;
% momentEqs determines what to do with the moment equations:
% 0 - derive moment equations from scratch
% 1 - load moment equations from existing file

N = 500;
% N is the total number of input simulations to be used by NMC and EKF

k_N = 100;
% k_N is the number of samples drawn by resampling (with replacement) of the full
% Monte Carlo sample at each time point (0 - no resampling)

dt = 0.5;
% dt is the time step of Monte Carlo simulation and EKF

Tend = 500;
% Tend: the end point of the simulation horizon

dq = 100;
% dq: sparsity factor for the Q matrix
%(Q matrix is calculated on the sparse time grid 0:dt*dq:Tend)

dr = 100;
% dr: sparsity factor for the R matrix
%(R matrix is calculated on the sparse time grid 0:dt*dr:Tend)

nRuns = 2;
% nRuns: number of NMC and EKF runs to be performed

%%% OUTPUT
% allEKF ~ EKF mean and variance trajectory estimates for all nRuns
% allMC ~ NMC mean and variance trajectory estimates for all nRuns

%----------------------------------------------------------------------------------------------------------------------------------------------------------
%%% Loading all the necessary files

% Load the model
modelNames = {'TCSallODEs_bi.txt','TCSallODEs_mono.txt','MAPK_feedback.txt'};
modeltxt = modelNames{nModel};
model=IQMmodel(modeltxt);
IQMmakeMEXmodel(model, 'model_MEX');
disp('Loading the model: done')

% Load reference moment trajectories from a large number of Monte Carlo
% simulations
referenceNames = {'momentsReference400000nz_TCSbi_Tend500.mat',...
	'momentsReference400000nz_TCSmono_Tend500.mat',...
	'momentsReference200000_MAPK_Tend500.mat'};
load(referenceNames{nModel});
disp('Loading the reference moment trajectories: done')

% Load or derive the moment equations
switch momentEqs
    case 1 % Load derived moment equations
        momentsEqsNames = {'momentsEquationsMEX_TCSbi.mat',...
            'momentsEquationsMEX_TCSmono.mat',...
            'momentsEquationsMEX_MAPK.mat'};
        load(momentsEqsNames{nModel});
    case 0 % Derive the moment equations from scratch
        %----------------------------------------------------------------------------------------------------------------------------------------------------------
        %%% Derivation of moment equations
        
        % Store IQM model components from the text file
        [states,ODEs,initialConditions] = IQMstates(model);
        [parameters,values] = IQMparameters(model);
        
        nstates = length(states);
        nparameters = length(parameters);
        
        % Create symbolic variables
        % x ~ state vector
        x = sym('x',[nstates,1],'real');
        ODEs = subs(str2sym(ODEs), states', x');
        % k ~ parameter vector
        k = sym('k',[nparameters,1],'real');
        ODEs = subs(ODEs, parameters', k');
        
        % Choose uncertain inputs (states and parameters)
        % ix ~ indices of states of interest
        ix = 1:nstates;
        
        % ik ~ indices of parameters of interest
        ik = [];
        
        inputs = [states(ix); parameters(ik)]; % inputs of interest
        symInputs = [x(ix); k(ik)]; % symbolic vector of inputs
        ninputs =  length(ix) + length(ik); % number of uncertain inputs
        
        ODEs = subs(ODEs, k(setdiff(1:end,ik))', values(setdiff(1:end,ik))');
        
        % Create symbolic variables for the moments of inputs
        m1 = sym('s',[ninputs,1],'real'); % first order moments of all inputs
        
        M2 = sym('s',[ninputs,ninputs],'real');
        m2 = diag(M2); % second order moments of all inputs
        
        moments = [m1; m2]; % moments of interest
        nmoments = length(moments);
        
        % Moment equations
        ODEs = expand(ODEs(ix)); % First order moment equations
        ODEs_X = expand(ODEs.*x(ix)); % Second order moment equations
        ODE = [ODEs; ODEs_X];
        
        % All power combinations of second order
        allPowers2 = powers(ninputs,2);
        [row2,col2] = find(allPowers2==2); % find second powers
        allPowers2(row2,:) = []; % remove the rows with second powers
        for i = 1 : size(allPowers2,1)
            allSymInter2(i,:) = prod(symInputs'.^allPowers2(i,:));
        end
        
        % Create list of existing power combinations of second order
        temp2 = cell(length(ODE),1);
        parfor i = 1:length(ODE)
            mns2 = nan(size(allPowers2,1),ninputs);
            [C2 T2] = coeffs(ODE(i));
            ii = 1;
            for j = 1:length(T2)
                im2 = find(T2(j)-allSymInter2==0);
                if ~isempty(im2)
                    ip2 = allPowers2(im2,:);
                    mns2(ii,:) = ip2;
                    ii = ii+1;
                else
                end
            end
            temp2{i} = mns2(all(~isnan(mns2),2),:);
            %temp2{i} = mns2(1:sum(~isnan(mns2(:,1))),:)';
        end
        existingPowers2 = cell2mat(temp2);
        existingPowers2 = unique(existingPowers2,'rows');
        
        % Create a list of existing interaction terms of second order
        % and of corresponding symbols
        for i = 1 : size(existingPowers2,1)
            symInter2(i,:) = prod(symInputs'.^existingPowers2(i,:));
            is = find(existingPowers2(i,:)==1);
            sInter2(i,1) = M2(is(1),is(2));
        end
        
        % All power combinations of third order
        allPowers3 = powers(ninputs,3);
        for i = 1 : size(allPowers3,1)
            allSymInter3(i,:) = prod(symInputs'.^allPowers3(i,:));
        end
        
        % Create list of existing power combinations of third order
        temp3 = cell(length(ODE),1);
        parfor i = 1:length(ODE)
            mns3 = nan(size(allPowers3,1),ninputs);
            [C3 T3] = coeffs(ODE(i));
            ii = 1;
            for j = 1:length(T3)
                im3 = find(T3(j)-allSymInter3==0);
                if ~isempty(im3)
                    ip3 = allPowers3(im3,:);
                    mns3(ii,:) = ip3;
                    ii = ii+1;
                else
                end
            end
            temp3{i} = mns3(all(~isnan(mns3),2),:);
        end
        existingPowers3 = cell2mat(temp3);
        existingPowers3 = unique(existingPowers3,'rows');
        
        % Create a list of existing interaction terms of third order
        % and of corresponding symbol
        format3 = "s%d_%d_%d";
        for i = 1 : size(existingPowers3,1)
            symInter3(i,:) = prod(symInputs'.^existingPowers3(i,:));
            is2 = find(existingPowers3(i,:)==2);
            is1 = find(existingPowers3(i,:)==1);
            if ~isempty(is2)
                sInter3(i,:) = str2sym(sprintf(format3,is2,is2,is1));
            else
                sInter3(i,:) = str2sym(sprintf(format3,is1(1),is1(2),is1(3)));
            end
        end
        
        highermoments = [sInter2; sInter3]; % all higher moments (~measurements)
        allExistingPowers = [existingPowers2; existingPowers3];
        
        % First order moment equations
        g1 = subs(ODEs, symInter3, sInter3);
        g1 = subs(g1, symInter2, sInter2);
        g1 = expand(subs(g1, x(ix).^2, m1.^2+m2));
        g1 = subs(g1, symInputs, m1);
        
        % Second order central moment equations
        g2 = subs(ODEs_X, symInter3, sInter3);
        g2 = subs(g2, symInter2, sInter2);
        g2 = expand(subs(g2, x(ix).^2, m1.^2+m2));
        g2 = subs(g2, symInputs, m1);
        g2 = simplify(2*g2 - 2*m1.*g1);
        
        % ODE moment system
        g = [g1; zeros(length(ik),1); g2; zeros(length(ik),1)];
        
        symJ = jacobian(g, moments); % Jacobian wrt the moments of interest
        symVarsJ = symvar(symJ); % List all the symbolic variables in symJ
        idMomentsJ = ismember(moments,symVarsJ); % Find the indices of moments in symJ
        idHighermomentsJ = ismember(highermoments,symVarsJ); % Find the indices of highermoments in symJ
        momentsJ = moments(idMomentsJ); % List the moments in symJ
        highermomentsJ = highermoments(idHighermomentsJ); % List the higher moments in symJ
        
        symG = jacobian(g, highermoments); % Jacobian wrt the higher moments (~ measurements)
        symVarsG = symvar(symG); % List all the symbolic variables in symG
        idMomentsG = ismember(moments,symVarsG); % Find the indices of moments in symG
        idHighermomentsG = ismember(highermoments,symVarsG); % Find the indices of highermoments in symG
        momentsG = moments(idMomentsG); % List the moments in symG
        highermomentsG = highermoments(idHighermomentsG); % List the higher moments in symG
        
end

% Create IQM model of moment equations
create_IQMtext_file(moments,g,highermoments);
moments_eqs = IQMmodel('moments.txt');
[moment_states,moment_ODEs,moment_initialConditions] = IQMstates(moments_eqs);
[moment_parameters,moment_values] = IQMparameters(moments_eqs);
IQMmakeMEXmodel(moments_eqs, 'moments_MEX'); % Uncomment to create a MEX file

disp('Moment equations derivation: done')

% Convert symbolic expression of ODEs into Matlab functions
syms t real
matlabFunction(g,'File','momentsODEfunction','Vars',{t,moments,highermoments});
matlabFunction(symJ,'File','jacobianJfunction','Vars',{momentsJ,highermomentsJ});
matlabFunction(symG,'File','jacobianGfunction','Vars',{momentsG,highermomentsG});

%----------------------------------------------------------------------------------------------------------------------------------------------------------
%%% Start of multiple runs
% Simulation time
Tsim = 0:dt:Tend;
T = Tend+1;

% Create matrices to store the output
allEKF = zeros(nmoments,length(Tsim),nRuns); % matrix to store all EKF runs
allMC = zeros(nmoments,length(Tsim),nRuns); % matrix to store all MC runs

for r = 1:nRuns
    
    % Construct random sample from independent log-normal distirbution
    switch nModel
        case {1,2} % Monte Carlo simulation for TCS systems
            %%% Sampling of initial conditions and parameters
            for i = 1:nstates
                if  i == 3
                    m = initialConditions(i);
                    v = 0.1*m^2;
                    %mu = log(m);
                    mu = log((m^2)/sqrt(v+m^2));
                    sigma = sqrt(log(v/(m^2)+1));
                    U(:,i) = lognrnd(mu,sigma,N,1);
                elseif ~isequal(initialConditions(i),0)
                    m = initialConditions(i);
                    v = 0.5*m^2;
                    %mu = log(m);
                    mu = log((m^2)/sqrt(v+m^2));
                    sigma = sqrt(log(v/(m^2)+1));
                    U(:,i) = lognrnd(mu,sigma,N,1);
                else
                    m = 0.001;
                    v = 0.5*m^2;
                    %mu = log(m);
                    mu = log((m^2)/sqrt(v+m^2));
                    sigma = sqrt(log(v/(m^2)+1));
                    U(:,i) = lognrnd(mu,sigma,N,1);
                end
            end
            
            W = zeros(N,length(ik));
            for i = 1:length(ik)
                m = values(ik(i));
                v = 0.5*m^2;
                mu = log((m^2)/sqrt(v+m^2));
                sigma = sqrt(log(v/(m^2)+1));
                W(:,i) = lognrnd(mu,sigma,N,1);
            end
            outParameters = repmat(W,1,1,length(Tsim));
            
            % Model simulation
            parfor n = 1:N
                out = IQMPsimulate('model_MEX', Tsim, U(n,:), parameters(ik), W(n,:));
                outStates(n,:,:) = out.statevalues';
                % 'number of runs' X 'number of states' X 'time'
            end
            
            % Trajectories of all uncertain inputs
            outInputs = cat(2,outStates,outParameters);
            
            
        case 3 % Monte Carlo simulation for the MAPK system
            %%% Sampling of initial conditions 
            U = zeros(N,nstates);
            for i = 1:nstates
                if ~isequal(initialConditions(i),0)
                    m = initialConditions(i);
                    v = 0.5*m^2;
                    mu = log((m^2)/sqrt(v+m^2));
                    sigma = sqrt(log(v/(m^2)+1));
                    U(:,i) = lognrnd(mu,sigma,N,1);
                else
                end
            end
            
            % Model simulation
            outStates = zeros(N,ninputs,length(Tsim));
            parfor n = 1:N
                out = IQMPsimulate('model_MEX', Tsim, U(n,:));
                outStates(n,:,:)=out.statevalues';
                % 'number of runs' X 'number of states' X 'time'
            end
            
            % Trajectories of all uncertain inputs
            outInputs = outStates;
            outParameters = [];
            
    end
    disp('Model simulation: done')
    
    %----------------------------------------------------------------------------------------------------------------------------------------------------------
    %%% Extended Kalman Filter algorithm
    % Resampling step
    switch k_N
        case 0 % no resampling
            d = N/length(Tsim);
            D = reshape(1:N, d, [])';
            trajectoryStates = zeros(N/length(Tsim),nstates,length(Tsim));
            for it = 1:length(Tsim)
                trajectoryStates(:,:,it) = outStates(D(it,:),:,it);
            end
            trajectoryInputs = trajectoryStates;
        otherwise
            d = k_N;
            trajectoryStates = zeros(d,nstates,length(Tsim));
            for it = 1:length(Tsim)
                [trajectoryStates(:,1,it), id1] = datasample(outStates(:,1,it),d,1);
                trajectoryStates(:,2:end,it) = outStates(id1,2:end,it);
            end
            trajectoryInputs = trajectoryStates;
    end
    
    % Compute moment estimates from the input Monte Carlo sample
    nhmoments = size(allExistingPowers,1);
    M = zeros(nmoments,length(Tsim)); % low-order moment 
    MM = zeros(nhmoments,length(Tsim)); % high-order moments
    M(1:ninputs,:) = permute((sum(trajectoryInputs)/d),[2 3 1]);
    M(ninputs+1:end,:) = permute(sum(bsxfun(@minus, trajectoryInputs, sum(trajectoryInputs)/d).^2)/(d-1),[2 3 1]);
    for ih = 1:nhmoments
        h1 = find(allExistingPowers(ih,:)==1);
        h2 = find(allExistingPowers(ih,:)==2);
        h = [h1 h2 h2];
        MM(ih,:) = sum(permute(prod(trajectoryInputs(:,h,:),2),[1 3 2]))/d;
    end
    
    %%% Initialization of Extended Kalman filter
    
    % Process noise: calculation of covariance matrix Q
    outQ = outInputs;
    N_Q = size(outQ,1);
    idq_sparse = unique([1:dq:length(Tsim) length(Tsim)]); %sparse time grid
    
    nhmoments = length(highermoments);
    qt = [];
    for i2 = 1:nhmoments
        for i1 = 1:i2
            qt = [qt;[i1 i2]];
        end
    end
    iqt = sub2ind([nhmoments nhmoments],qt(:,1),qt(:,2));
    
    Q_sparse = zeros(nhmoments,nhmoments,length(idq_sparse));
    Q = zeros(nhmoments,nhmoments,length(Tsim));
    for ii = 1:length(iqt)
        [row, col] = ind2sub([nhmoments nhmoments], iqt(ii));
        ir1 = find(allExistingPowers(row,:)==1);
        ir2 = find(allExistingPowers(row,:)==2);
        ir = [ir1 ir2 ir2];
        ic1 = find(allExistingPowers(col,:)==1);
        ic2 = find(allExistingPowers(col,:)==2);
        ic = [ic1 ic2 ic2];
        iall = [ir ic];
        dummy1 = squeeze(prod(outQ(:,iall,idq_sparse),2));
        dummy2 = squeeze(prod(outQ(:,ir,idq_sparse),2));
        dummy3 = squeeze(prod(outQ(:,ic,idq_sparse),2));
        Q_sparse(row,col,:) = (sum(dummy1)/(N_Q-1) - (sum(dummy2).*sum(dummy3))/(N_Q*(N_Q-1)))/d;
    end
    
    % Observation noise: calculation of covariance matrix R
    outR = outInputs;
    N_R = size(outR,1);
    nr = ninputs;
    idr_sparse = unique([1:dr:length(Tsim) length(Tsim)]); %sparse time grid
    
    rt = [];
    for i2 = 1:nr
        for i1 = 1:i2
            rt = [rt;[i1 i2]];
        end
    end
    irt = sub2ind([nr nr],rt(:,1),rt(:,2));
    
    R_sparse = zeros(nmoments, nmoments, length(idr_sparse));
    R = zeros(2*nr, 2*nr, length(Tsim));
    for ii = 1:length(irt)
        [row, col] = ind2sub([nr nr], irt(ii));
        s10 = sum(outR(:,row,idr_sparse));
        s01 = sum(outR(:,col,idr_sparse));
        s11 = sum(outR(:,row,idr_sparse).*outR(:,col,idr_sparse));
        s20 = sum(outR(:,row,idr_sparse).^2);
        s02 = sum(outR(:,col,idr_sparse).^2);
        s22 = sum(outR(:,row,idr_sparse).^2.*outR(:,col,idr_sparse).^2);
        s21 = sum(outR(:,row,idr_sparse).^2.*outR(:,col,idr_sparse));
        s12 = sum(outR(:,row,idr_sparse).*outR(:,col,idr_sparse).^2);
        R_sparse(row,col,:) = (s11/(N_R-1)-(s10.*s01)/(N_R*(N_R-1)))/d;
        R_sparse(row+nr,col+nr,:) = 1/d*(s22/N_R) ...
            -2/d*(s10.*s12-s22)/(N_R*(N_R-1)) ...
            -2/d*(s01.*s21-s22)/(N_R*(N_R-1)) ...
            -1/d*(s20.*s02-s22)/(N_R*(N_R-1)) ...
            +2/d*(s01.^2.*s20-s02.*s20-2*s01.*s21+2*s22)/(N_R*(N_R-1)*(N_R-2)) ...
            +2/d*(s10.^2.*s02-s02.*s20-2*s10.*s12+2*s22)/(N_R*(N_R-1)*(N_R-2)) ...
            +2/(d*(d-1))*(s11.^2-s22)/(N_R*(N_R-1)) ...
            +4*(d-2)/(d*(d-1))*(s01.*s10.*s11-s11.^2-s10.*s12-s01.*s21+2*s22)/(N_R*(N_R-1)*(N_R-2)) ...
            -2*(2*d-3)/(d*(d-1))*(s01.^2.*s10.^2-s02.*s10.^2-s20.*s01.^2-4*s01.*s10.*s11 ...
            +2*s11.^2+4*s10.*s12+4*s01.*s21+s02.*s20-6*s22)/(N_R*(N_R-1)*(N_R-2)*(N_R-3));
        R_sparse(row,col+nr,:) = 1/d*(2*s01.^2.*s10-N_R*s02.*s10-2*N_R*s01.*s11 ...
            +N_R^2.*s12)/(N_R*(N_R-1)*(N_R-2));
        R_sparse(col,row+nr,:) = 1/d*(2*s10.^2.*s01-N_R*s20.*s01-2*N_R*s10.*s11 ...
            +N_R^2.*s21)/(N_R*(N_R-1)*(N_R-2));
    end
    R(:,:,1) = R_sparse(:,:,1);
    H = eye(nmoments); %output matrix (selects which low-order moments are used in filtering)
    
    mu_ = [momentsReference(1:nmoments,1) zeros(nmoments,length(Tsim)-1)]; %vector to store prior mean estimates
    P_(:,:,1) = 1e-10*eye(nmoments);  % matrix to store prior covariance estimates
    
    muhat = [momentsReference(1:nmoments,1) zeros(nmoments,length(Tsim)-1)]; %vector to store posterior mean estimates
    P(:,:,1) = 1e-10*eye(nmoments); % matrix to store posterior covariance estimates
    
    %%% Kalman Filter prediction and update
    for kk = 2:length(Tsim)
        
        % Jacobian matrices used in the linearization
        J(:,:,kk-1) = jacobianJfunction(muhat(idMomentsJ,kk-1), MM(idHighermomentsJ,kk-1));
        G = jacobianGfunction(muhat(idMomentsG,kk-1), MM(idHighermomentsG,kk-1));
        
        % Solving ODEs for the moments
        mu_out = IQMPsimulate('moments_MEX', [Tsim(kk-1) Tsim(kk)], muhat(:,kk-1), moment_parameters, MM(:,kk-1));
        mu_(:,kk) = mu_out.statevalues(end,:)';
        
        % Linear interpolation of Q matrix
        tq = find(kk-1 >= idq_sparse, 1, 'last');
        dqTsim = Tsim(idq_sparse(tq+1)) - Tsim(idq_sparse(tq));
        dummyQ = Q_sparse(:,:,tq) + (Q_sparse(:,:,tq+1)-Q_sparse(:,:,tq))/dqTsim*dt*(kk-1-idq_sparse(tq));
        Q(:,:,kk-1) = dummyQ + triu(dummyQ,1)' ;
        
        % Linear interpolation of R matrix
        if kk~= length(Tsim)
            tr = find(kk >= idr_sparse, 1, 'last');
            drTsim = Tsim(idr_sparse(tr+1)) - Tsim(idr_sparse(tr));
            dummyR = R_sparse(:,:,tr) + (R_sparse(:,:,tr+1)-R_sparse(:,:,tr))/drTsim*dt*(kk-idr_sparse(tr));
            R(:,:,kk) = dummyR + triu(dummyR,1)';
        else
            R(:,:,kk) = R_sparse(:,:,end) + triu(R_sparse(:,:,end),1)';
        end
        
        % Solving ODEs for the covariance matrix
        P_(:,:,kk) = odefcnG_CT(dt,P(:,:,kk-1),J(:,:,kk-1),Q(:,:,kk-1),G);
        
        % Calculate Kalman gain
        K = P_(:,:,kk)*H'*pinv(H*P_(:,:,kk)*H' + R(:,:,kk));
        
        % Update
        mu = mu_(:,kk) + K*H*(M(:,kk)-mu_(:,kk)); %
        P(:,:,kk) = (eye(size(P,1))-K*H)*P_(:,:,kk);
        
        % Store the updated estimates
        muhat(:,kk) = mu;
        
    end
    
    allEKF(:,:,r) = muhat;
    disp('Extended Kalman Filter estimated moment trajectories for a single run: done')
    
    %----------------------------------------------------------------------------------------------------------------------------------------------------------
    %%% Naive Monte Carlo (NMC) based moment trajectories
    MC = zeros(nmoments,length(Tsim)); % moments of interest
    for it = 1:length(Tsim)
        dummyMean = sum(outInputs(:,:,it))/N; % mean
        dummyVar = sum(bsxfun(@minus, outInputs(:,:,it), dummyMean).^2)/(N-1); % variance
        MC(:,it) = [dummyMean'; dummyVar'];
    end
    
    allMC(:,:,r) = MC;
    disp('Naive Monte Carlo estimated moment trajectories for a single run: done')
    
end

save(['EKF_MC_output_N' num2str(N) '_Tend' num2str(T) '_dt' num2str(dt) '_dq'  num2str(dq) '_dr' num2str(dr) '_X' num2str(nRuns) '.mat'], ...
    'allEKF', 'allMC','Tsim')
